# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import paddle.fluid as fluid


class ROIPooling(object):
    def __init__(self, pooled_height, pooled_width, spatial_scale):
        self.roipooling_layer_attrs = {
            "pooled_height": pooled_height,
            "pooled_width": pooled_width,
            "spatial_scale": spatial_scale
        }

    def __call__(self, x0, x1):
        slice_x1 = paddle.slice(input=x1, axes=[1], starts=[1], ends=[5])
        out = fluid.layers.roi_pool(
            input=x0, rois=slice_x1, **self.roipooling_layer_attrs)
        return out



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import paddle.fluid as fluid


class PriorBox(object):
    def __init__(self, min_sizes, max_sizes, aspect_ratios, variance, flip,
                 clip, steps, offset, min_max_aspect_ratios_order):
        self.priorbox_layer_attrs = {
            "min_sizes": min_sizes,
            "max_sizes": max_sizes,
            "aspect_ratios": aspect_ratios,
            "variance": variance,
            "flip": flip,
            "clip": clip,
            "steps": steps,
            "offset": offset,
            "min_max_aspect_ratios_order": min_max_aspect_ratios_order
        }

    def __call__(self, x0, x1):
        box, var = fluid.layers.prior_box(
            input=x0, image=x1, **self.priorbox_layer_attrs)
        box = paddle.reshape(x=box, shape=[1, 1, -1])
        var = paddle.reshape(x=var, shape=[1, 1, -1])
        out = paddle.concat(x=[box, var], axis=1)
        return out



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle


class LocalResponseNorm(object):
    def __init__(self, size, alpha=1e-4, beta=0.75, k=1.):
        self.size = size
        self.alpha = alpha
        self.beta = beta
        self.k = k

    def __call__(self, x):
        sizes = x.shape
        dim = len(sizes)
        if dim < 3:
            raise ValueError(
                'Expected 3D or higher dimensionality input, but got {} dimensions'.
                format(dim))
        div = paddle.unsqueeze(paddle.multiply(x, x), axis=1)
        pad4d_shape = [self.size // 2, (self.size - 1) // 2, 0, 0]
        pool2d_shape = (1, self.size)
        pad5d_shape = [self.size // 2, (self.size - 1) // 2, 0, 0, 0, 0]
        pool3d_shape = (1, 1, self.size)

        if dim == 3:
            div = paddle.nn.functional.pad(div, pad=pad4d_shape)
            div = paddle.nn.functional.avg_pool2d(
                div, kernel_size=pool2d_shape, stride=1)
            div = paddle.squeeze(div, axis=1)
        else:
            tmp = paddle.unsqueeze(x, axis=1)
            reshape_shape = paddle.shape(tmp)
            new_reshape_shape = paddle.cast(reshape_shape, "float32")
            index = paddle.full(shape=[1], fill_value=-2, dtype="int32")
            value = paddle.full(shape=[1], fill_value=-1, dtype="float32")
            new_reshape_shape = paddle.scatter(new_reshape_shape, index, value)
            new_reshape_shape = paddle.cast(new_reshape_shape, "int32")
            div = paddle.reshape(div, shape=reshape_shape)
            div = paddle.nn.functional.pad(div,
                                           pad=pad5d_shape,
                                           data_format='NCDHW')
            div = paddle.nn.functional.avg_pool3d(
                div, kernel_size=pool3d_shape, stride=1)
            div = paddle.reshape(paddle.squeeze(div, axis=1), sizes)

        div = paddle.scale(div, scale=self.alpha, bias=self.k)
        div = paddle.pow(div, self.beta)
        res = paddle.divide(x, div)
        return res



# Copyright (c) 2019  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from x2paddle.op_mapper.onnx2paddle.opset9 import OpSet9
from x2paddle.decoder.onnx_decoder import ONNXGraphNode
from x2paddle.core.program import PaddleGraph


class ONNXOpMapper():
    def __init__(self, decoder):
        self.support_op_sets = [9, ]
        self.default_op_set = 9
        self.graph = decoder.graph
        self.paddle_graph = PaddleGraph(parent_layer=None, source_type="onnx")
        self.paddle_graph.outputs = self.graph.output_nodes
        self.opset = self.create_opset(decoder)
        if not self.op_checker():
            raise Exception("Model is not supported yet.")

        print("Total nodes: {}".format(
            sum([
                isinstance(node, ONNXGraphNode)
                for name, node in self.graph.node_map.items()
            ])))
        print("Nodes converting ...")
        for i, node_name in enumerate(self.graph.topo_sort):
            sys.stderr.write("\rConverting node {} ...     ".format(i + 1))
            node = self.graph.get_node(node_name)
            op = node.layer_type
            if hasattr(self.opset, op):
                func = getattr(self.opset, op)
                func(node)
            elif op in self.opset.directly_map_ops:
                self.opset.directly_map(node)
            elif op in self.opset.elementwise_ops:
                self.opset.elementwise_map(node)
        print("\nNodes converted.")
        self.paddle_graph.set_name(self.graph.graph_name)
        self.paddle_graph.set_parameters(self.opset.weights)
        self.paddle_graph.set_inputs_info(self.opset.inputs_info)

    def op_checker(self):
        unsupported_ops = set()
        for node_name in self.graph.topo_sort:
            node = self.graph.get_node(node_name)
            op = node.layer_type
            if not hasattr(self.opset, op) and \
                op not in self.opset.directly_map_ops and \
                op not in self.opset.elementwise_ops:
                unsupported_ops.add(op)
        if len(unsupported_ops) == 0:
            return True
        else:
            if len(unsupported_ops) > 0:
                print("\n========= {} OPs are not supported yet ===========".
                      format(len(unsupported_ops)))
            for op in unsupported_ops:
                print("========== {} ============".format(op))
            return False

    def create_opset(self, decoder):
        run_op_set = self.default_op_set
        opset = ''
        if decoder.op_set in self.support_op_sets:
            opset = 'OpSet' + str(decoder.op_set)
        elif decoder.op_set < self.default_op_set:
            opset = 'OpSet' + str(self.default_op_set)
        else:
            for op_set in self.support_op_sets:
                if decoder.op_set > op_set:
                    run_op_set = op_set
                else:
                    break
            opset = 'OpSet' + str(run_op_set)
        print(
            'Now, onnx2paddle support convert onnx model opset_verison {},'
            'opset_verison of your onnx model is {}, automatically treated as op_set: {}.'
            .format(self.support_op_sets, decoder.op_set, run_op_set))
        return eval(opset)(decoder, self.paddle_graph)



# -*- coding:UTF-8 -*-
# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle


def string(param):
    """ 生成字符串。
    """
    return "\'{}\'".format(param)


def check_version():
    version = paddle.__version__
    v0, v1, v2 = version.split('.')
    if not ((v0 == '0' and v1 == '0' and v2 == '0') or
            (int(v0) >= 2 and int(v1) >= 1)):
        return False
    else:
        return True


class PaddleDtypes():
    def __init__(self, is_new_version=True):
        if is_new_version:
            self.t_float16 = paddle.float16
            self.t_float32 = paddle.float32
            self.t_float64 = paddle.float64
            self.t_uint8 = paddle.uint8
            self.t_int8 = paddle.int8
            self.t_int16 = paddle.int16
            self.t_int32 = paddle.int32
            self.t_int64 = paddle.int64
            self.t_bool = paddle.bool
        else:
            from paddle.fluid.core import VarDesc
            self.t_float16 = VarDesc.VarType.FP16
            self.t_float32 = VarDesc.VarType.FP32
            self.t_float64 = VarDesc.VarType.FP64
            self.t_uint8 = VarDesc.VarType.UINT8
            self.t_int8 = VarDesc.VarType.INT8
            self.t_int16 = VarDesc.VarType.INT16
            self.t_int32 = VarDesc.VarType.INT32
            self.t_int64 = VarDesc.VarType.INT64
            self.t_bool = VarDesc.VarType.BOOL


is_new_version = check_version()
paddle_dtypes = PaddleDtypes(is_new_version)
